! LOGSTEP Test Module: driver_batch_sep_dc.F90
!
! * This program is written by Yusuke HIROTA.
!
! # License (MIT License)
! Copyright 2018 RIKEN.
!
! Permission is hereby granted, free of charge, to any person obtaining
! a copy of this software and associated documentation files (the
! "Software"), to deal in the Software without restriction, including
! without limitation the rights to use, copy, modify, merge, publish,
! distribute, sublicense, and/or sell copies of the Software, and to
! permit persons to whom the Software is furnished to do so, subject to
! the following conditions:
!
! The above copyright notice and this permission notice shall be
! included in all copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
! MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
! IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
! CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
! TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
! SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!
program main
  use batch_sep_dc_timing
  implicit none
  integer, parameter :: ld_offset = 4
  !
  integer :: num_iter
  integer :: dummy_size
  character :: matrix_type 
  ! "F": Inverse of the Frank matrix,
  ! "R": Random matrix
  ! "U": Unbalanced matri, diag(F, R) + [-1, 1; 1, -1]. Assume n/2 is the first split point.
  ! "M": Empirically midpoint between 'F' and 'R'
  integer :: n
  character :: method
  ! "D": DSTEDC
  ! "P": PLASMA DSTEDC
  ! "B": Batch DC
  logical :: check(3) ! Eigenvalues, relative residual
  integer :: npcores ! Used only for PLASMA
  !
  real(kind(0.0d0)), allocatable :: d(:), e(:)
  real(kind(0.0d0)), allocatable :: x(:, :)
  real(kind(0.0d0)), allocatable :: dummy(:)
  integer :: ldx
  real(kind(0.0d0)) :: etime
  type(batch_sep_dc_timing_list) :: batch_sep_dc_timer
  real(kind(0.0d0)) :: accuracy(3)
  integer :: i
  integer :: info
  !
  logical :: lsame

  !> Read test settings
  read(*, *) num_iter
  read(*, *) dummy_size
  read(*, *) matrix_type
  read(*, *) n
  read(*, *) method
  read(*, *) check(1)
  read(*, *) check(2)
  read(*, *) check(3)
  read(*, *) npcores
  print *, "<INFO> num_iter, dummy_size, matrix_type, n = ", &
           num_iter, dummy_size, matrix_type, n
  print *, "<INFO> method, check(1:3) = ", method, check(1:3)
  print *, "<INFO> npcores = ", npcores

  !> Check input parameters
  if ((num_iter < 1) .or. (dummy_size < 0)) then
    call abort()
  endif
  if (.not. (lsame(matrix_type, 'F') .or. lsame(matrix_type, 'R') .or. &
             lsame(matrix_type, 'U') .or. lsame(matrix_type, 'M'))) then
    call abort()
  endif
  if (.not. (lsame(method, 'D') .or. lsame(method, 'P') .or. lsame(method, 'B'))) then
    call abort()
  endif
  if (n < 1) then
    call abort()
  endif
  if (npcores < 1) then
    call abort()
  endif
  !> Calculate the matrix total dimensions and allocate spaces.
  ldx = n + ld_offset
  allocate(d(n), e(n - 1), x(ldx, n), dummy(dummy_size), stat = info)
  if (info /= 0) call abort()

  !> Execute benchmark.
  call bench_sep_dc(num_iter, dummy_size, dummy, matrix_type, method, check, npcores, &
                    n, d, e, x, ldx, etime, batch_sep_dc_timer, accuracy)
  if (check(1)) print *, "accuracy(eigenvalue) = ", accuracy(1)
  if (check(2)) print *, "accuracy(relative-residual) = ", accuracy(2)
  if (check(3)) print *, "accuracy(relative-orthogonality) = ", accuracy(3)
  print *, "etime [sec] = ", etime
  if (lsame(method, 'B')) then
    print *, "batch_sep_dc dstedc [sec] ", batch_sep_dc_timer%t_dstedc_batch
    print *, "batch_sep_dc dlaed0 [sec] ", batch_sep_dc_timer%t_dlaed0_batch
    print *, "batch_sep_dc dlaed0b [sec] ", batch_sep_dc_timer%t_dlaed0b_batch
    print *, "batch_sep_dc sum(dlaed1a) [sec] ", sum(batch_sep_dc_timer%t_dlaed1a_batch)
    print *, "batch_sep_dc sum(dlaed1b) [sec] ", sum(batch_sep_dc_timer%t_dlaed1b_batch)
    print *, "batch_sep_dc sum(dlaed2) [sec] ", sum(batch_sep_dc_timer%t_dlaed2)
    print *, "batch_sep_dc sum(dlaed3) [sec] ", sum(batch_sep_dc_timer%t_dlaed3_batch)
    print *, "batch_sep_dc sum(dlaed4) [sec] ", sum(batch_sep_dc_timer%t_dlaed4)
    print *, "batch_sep_dc sum(batch_dgemm) [sec] ", sum(batch_sep_dc_timer%t_batch_dgemm)
    do i = 1, batch_sep_dc_timing_max_level
      print *, "i, batch_sep_dc {dlaed{1a,1b,2,3,4},batch_dgemm}(i) [sec] ", i, &
               batch_sep_dc_timer%t_dlaed1a_batch(i), batch_sep_dc_timer%t_dlaed1b_batch(i), &
               batch_sep_dc_timer%t_dlaed2(i), &
               batch_sep_dc_timer%t_dlaed3_batch(i), batch_sep_dc_timer%t_dlaed4(i), &
               batch_sep_dc_timer%t_batch_dgemm(i)
    enddo
    do i = 1, batch_sep_dc_timing_max_level
      print *, "i, general purpose timer(i) [sec] ", i, batch_sep_dc_timer%t_general(i)
    enddo
  endif

  deallocate(d, e, x, dummy)
  stop
end


!> Working space will be allocated in the subroutine.
subroutine bench_sep_dc(num_iter, dummy_size, dummy, matrix_type, method, check, npcores, &
                        n, d, e, x, ldx, etime, batch_sep_dc_timer, accuracy)
  use omp_lib
  use batch_sep_dc_timing
  use plasma
  implicit none
  integer, parameter :: rseed = 100
  !
  integer, intent(in) :: num_iter
  integer, intent(in) :: dummy_size, n
  real(kind(0.0d0)), intent(out) :: dummy(*)
  character, intent(in) :: matrix_type
  logical, intent(in) :: check(*)
  character, intent(in) :: method
  integer, intent(in) :: npcores
  real(kind(0.0d0)), intent(out) :: accuracy(3)
  real(kind(0.0d0)), intent(out) :: d(*), e(*)
  real(kind(0.0d0)), intent(out) :: x(*)
  integer, intent(in) :: ldx
  real(kind(0.0d0)), intent(out) :: etime
  type(batch_sep_dc_timing_list), intent(out) :: batch_sep_dc_timer
  !
  integer :: iter
  real(kind(0.0d0)), allocatable :: work(:)
  real(kind(0.0d0)), allocatable :: s2(:)
  integer, allocatable :: iwork(:)
  integer, allocatable :: omp_iwork(:)
  integer :: lwork, liwork
  real(kind(0.0d0)) :: rarray(1)
  integer :: iarray(1)
  integer :: info
  real(kind(0.0d0)), allocatable :: d_orig(:), e_orig(:)
  !
  integer, allocatable :: seed(:)
  integer :: seed_size
  logical :: lsame
  !
  real(kind(0.0d0)) :: t1, t2
  type(batch_sep_dc_timing_list) :: this_batch_sep_dc_timer
  !
  logical, parameter :: p_autotuning = .false.
  ! At PLASMA 2.8.0 plasma_dstedc, the parameters are irrelevant to the performance. Thus we set the autotuning values of DSTYEVD.
  integer, parameter :: p_nb = 120, p_ib = 20

  !> Initialization
  call random_seed(size = seed_size)
  allocate(seed(seed_size), stat = info)
  if (info /= 0) call abort()
  seed(1 : seed_size) = rseed
  !
  if (lsame(method , "P")) then
    call plasma_init(npcores, info)
    if (p_autotuning) then
      call plasma_enable(PLASMA_AUTOTUNING, info)
    else
      call plasma_disable(PLASMA_AUTOTUNING, info)
      call plasma_set(PLASMA_TILE_SIZE, p_nb, info)
      call plasma_set(PLASMA_INNER_BLOCK_SIZE, p_ib, info)
    endif
    call plasma_set(PLASMA_SCHEDULING_MODE, PLASMA_DYNAMIC_SCHEDULING, info)
  endif

  !> Workspace allocation
  if (lsame(method, "D")) then
    call dstedc('I', n, d, e, x, ldx, rarray, -1, iarray, -1, info)
    lwork = int(rarray(1)) + 1
    liwork = iarray(1)
  else if (lsame(method, "P")) then
    ! At PLASMA 2.8.0 plasma_dstedc, workspace is allocated in the PLSAMA routine.
    lwork = 1
    liwork = 1
  else if (lsame(method, "B")) then
    call batch_sep_dc(n, d, e, x, ldx, rarray, -1, iarray, -1, rarray, iarray, info, this_batch_sep_dc_timer)
    lwork = int(rarray(1)) + 1
    liwork = iarray(1)
  else
    call abort()
  endif
  if (check(2) .or. check(3)) then
    lwork = max(lwork, n * n) ! work is also used for accuracy check.
  else
    lwork = max(lwork, 1)
  endif
  allocate(work(lwork), iwork(liwork), s2(n * n), omp_iwork(4 * n), d_orig(n), e_orig(n - 1), stat = info)
  if (info /= 0) call abort()

  !> Benchmark
  etime = huge(0.0d0) ! Max value of real(kind(0.0d0).
  do iter = 0, num_iter ! iter == 0 is for warm up.
    ! Reset breakdown timer
    call batch_sep_dc_timing_reset(this_batch_sep_dc_timer)

    ! Update the matrix
    if (lsame(matrix_type, 'F')) then
      d(1) = 1.0d0
      d(2 : n) = 2.0d0
      e(1 : n - 1) = -1.0d0
    else if(lsame(matrix_type, 'R')) then
      ! Most of the eigenpairs may be deflated.
      call random_seed(put = seed)
      call random_number(d(1 : n))
      call random_number(e(1 : n - 1))
    else if(lsame(matrix_type, 'U')) then
      ! Unbalanced, diag(F, R) + [-1, 1; 1, -1]
      ! Assume n/2 is the first split point.
      call random_seed(put = seed)
      call random_number(d(1 : n / 2))
      call random_number(e(1 : n / 2 - 1))
      d(n / 2) = d(n / 2) - 1.0d0
      e(n / 2) = 1.0d0
      d(n / 2 : n) = 2.0d0
      e(n / 2 + 1 : n - 1) = -1.0d0
    else if(lsame(matrix_type, 'M')) then
      ! Empirically midpoint between 'F' and 'R'
      call random_seed(put = seed)
      call random_number(d(1 : n))
      e(1 : n - 1) = -1.0d0
    else
      call abort()
    endif
    d_orig(1 : n) = d(1 : n)
    e_orig(1 : n - 1) = e(1 : n - 1)

    ! Flush cache
    call random_number(dummy(1 : dummy_size))

    ! Main execution
    t1 = omp_get_wtime()
    if (lsame(method, "D")) then
      call dstedc('I', n, d, e, x, ldx, work, lwork, iwork, liwork, info)
    else if (lsame(method, "P")) then
      call plasma_dstedc(PlasmaIVec, n, d, e, x, ldx, info)
    else if (lsame(method, "B")) then
      call batch_sep_dc(n, d, e, x, ldx, work, lwork, iwork, liwork, s2, omp_iwork, info, this_batch_sep_dc_timer)
    else
      call abort()
    endif
    t2 = omp_get_wtime()
    if (info /= 0) then
      print *, "Main execution failed. info = ", info
      call abort()
    endif
    ! Check timing.
    if (iter /= 0) then
      if (t2 - t1 < etime) then
        etime = t2 - t1
        if (lsame(method, "B")) then
          batch_sep_dc_timer = this_batch_sep_dc_timer
        endif
        ! Accuracy check.
        if (check(1)) call get_max_relative_error_sep_eigenvalue(matrix_type, n, d, accuracy(1))
        if (check(2)) call get_relative_residual(n, d_orig, e_orig, d, x, ldx, work, n, accuracy(2))
        if (check(3)) call get_relative_orthogonality(n, x, ldx, work, n, accuracy(3))
      endif
    endif
  enddo

  if (lsame(method, "P")) then
    call plasma_finalize(info)
  endif
  !> Workspace deallocation
  deallocate(work, iwork, s2, d_orig, e_orig, seed)
end


!> set max_relative_error : if true eigenvalues are known max_i | d(i) - d_true(i) |,
!!                          otherwise -1.0d0.
subroutine get_max_relative_error_sep_eigenvalue(matrix_type, n, d, max_relative_error)
  implicit none
  character, intent(in) :: matrix_type
  integer, intent(in) :: n
  real(kind(0.0d0)), intent(in) :: d(*)
  real(kind(0.0d0)), intent(out) :: max_relative_error
  !
  integer :: i
  real(kind(0.0d0)) :: this_true_eigenvalue
  logical :: lsame

  if (lsame(matrix_type, 'F')) then
    max_relative_error = 0.0d0
    do i = 1, n
      this_true_eigenvalue = 4.0d0 * sin((4.0d0 * atan(1.0d0) * (2 * i - 1)) / (4 * n + 2)) ** 2.0d0
      max_relative_error = max(max_relative_error, &
                               abs((d(i) - this_true_eigenvalue) / this_true_eigenvalue))
    enddo
  else
    max_relative_error = -1.0d0
  endif
end


!> set relative_residual : ||A X - X D||_F / ||A||_F
subroutine get_relative_residual(n, d_orig, e_orig, d, x, ldx, w, ldw, relative_residual)
  implicit none
  integer, intent(in) :: n, ldx, ldw
  real(kind(0.0d0)), intent(in) :: d_orig(*), e_orig(*), d(*), x(ldx, *)
  real(kind(0.0d0)), intent(out) :: w(ldw, *)
  real(kind(0.0d0)), intent(out) :: relative_residual
  !
  integer :: i
  real(kind(0.0d0)) :: afnorm
  real(kind(0.0d0)) :: rarray(1)
  real(kind(0.0d0)) :: dlange

  !> ||A||_F
  afnorm = dlange('F', n, 1, d_orig, 1, rarray) ** 2.0d0
  afnorm = afnorm + 2.0d0 * dlange('F', n - 1, 1, e_orig, 1, rarray) ** 2.0d0
  afnorm = sqrt(afnorm)
  !> W <-- - X D
  call dlaset('A', n, n, 0.0d0, 0.0d0, w, ldw)
  !$omp parallel
  !$omp do
  do i = 1, n
    call daxpy(n, -d(i), x(1, i), 1, w(1, i), 1)
  enddo
  !$omp end do

  !> W <-- A X + W = A X - X D
  !$omp do
  do i = 1, n
    if (i /= 1) then
      call daxpy(n, e_orig(i - 1), x(i - 1, 1), ldx, w(i, 1), ldw)
    endif
    call daxpy(n, d_orig(i), x(i, 1), ldx, w(i, 1), ldw)
    if (i /= n) then
      call daxpy(n, e_orig(i), x(i + 1, 1), ldx, w(i, 1), ldw)
    endif
  enddo
  !$omp end do
  !$omp end parallel

  !> Get the norm of W
  relative_residual = dlange('F', n, n, w, ldw, rarray) / afnorm
end


!> set relative_orthogonality : ||X^T X - I||_F / sqrt(n)
subroutine get_relative_orthogonality(n, x, ldx, w, ldw, relative_orthogonality)
  implicit none
  integer, intent(in) :: n, ldx, ldw
  real(kind(0.0d0)), intent(in) :: x(*)
  real(kind(0.0d0)), intent(out) :: w(*)
  real(kind(0.0d0)), intent(out) :: relative_orthogonality
  !
  real(kind(0.0d0)) :: dlange
  real(kind(0.0d0)) :: rarray(1)

  call dlaset('A', n, n, 0.0d0, 1.0d0, w, ldw)
  call dgemm('T', 'N', n, n, n, 1.0d0, x, ldx, x, ldx, -1.0d0, w, ldw)
  relative_orthogonality = dlange('F', n, n, w, ldw, rarray)
end

